@@ -34,6 +34,7 @@
 
 #define GOTD_PROG_NOTIFY_EMAIL	got-notify-email
 #define GOTD_PROG_NOTIFY_HTTP	got-notify-http
+#define GOTD_PROG_GOTSYS	gotsys
 
 #define GOTD_PATH_PROG_NOTIFY_EMAIL \
 	GOTD_STRINGVAL(GOT_LIBEXECDIR) "/" \
@@ -41,6 +42,9 @@
 #define GOTD_PATH_PROG_NOTIFY_HTTP \
 	GOTD_STRINGVAL(GOT_LIBEXECDIR) "/" \
 	GOTD_STRINGVAL(GOTD_PROG_NOTIFY_HTTP)
+#define GOTD_PATH_PROG_GOTSYS \
+	GOTD_STRINGVAL(GOT_LIBEXECDIR) "/" \
+	GOTD_STRINGVAL(GOTD_PROG_GOTSYS)
 
 #define GOTD_MAXCLIENTS		1024
 #define GOTD_MAX_CONN_PER_UID	4
@@ -54,16 +58,19 @@
 #define GOTD_CLIENT_TABLE_SIZE (GOTD_MAXCLIENTS * 4)
 
 enum gotd_procid {
-	PROC_GOTD	= 0,
-	PROC_LISTEN,
-	PROC_AUTH,
-	PROC_SESSION_READ,
-	PROC_SESSION_WRITE,
-	PROC_REPO_READ,
-	PROC_REPO_WRITE,
-	PROC_GITWRAPPER,
-	PROC_NOTIFY,
-	PROC_MAX,
+	GOTD_PROC_GOTD	= 0,
+	GOTD_PROC_LISTEN,
+	GOTD_PROC_AUTH,
+	GOTD_PROC_SESSION_READ,
+	GOTD_PROC_SESSION_WRITE,
+	GOTD_PROC_REPO_READ,
+	GOTD_PROC_REPO_WRITE,
+	GOTD_PROC_GITWRAPPER,
+	GOTD_PROC_NOTIFY,
+	GOTD_PROC_GOTSYS,
+	GOTD_PROC_RELOAD,
+	GOTD_PROC_GOTCTL,
+	GOTD_PROC_MAX,
 };
 
 struct gotd_imsgev {
@@ -129,11 +136,16 @@ struct gotd_repo {
 
 	struct gotd_access_rule_list rules;
 	struct got_pathlist_head protected_tag_namespaces;
+	size_t nprotected_tag_namespaces;
 	struct got_pathlist_head protected_branch_namespaces;
+	size_t nprotected_branch_namespaces;
 	struct got_pathlist_head protected_branches;
+	size_t nprotected_branches;
 
 	struct got_pathlist_head notification_refs;
+	size_t num_notification_refs;
 	struct got_pathlist_head notification_ref_namespaces;
+	size_t num_notification_ref_namespaces;
 	struct gotd_notification_targets notification_targets;
 };
 TAILQ_HEAD(gotd_repolist, gotd_repo);
@@ -165,12 +177,14 @@ struct gotd {
 	int nrepos;
 	struct gotd_child_proc *listen_proc;
 	struct gotd_child_proc *notify_proc;
+	struct gotd_child_proc *reload_proc;
 	int notifications_enabled;
 	struct timeval request_timeout;
 	struct timeval auth_timeout;
 	struct gotd_uid_connection_limit *connection_limits;
 	size_t nconnection_limits;
 	struct gotd_secrets *secrets;
+	char *default_sender;
 
 	char *argv0;
 	const char *confpath;
@@ -187,6 +201,7 @@ enum gotd_imsg_type {
 	GOTD_IMSG_INFO_REPO,
 	GOTD_IMSG_INFO_CLIENT,
 	GOTD_IMSG_STOP,
+	GOTD_IMSG_RELOAD,
 
 	/* Request a list of references. */
 	GOTD_IMSG_LIST_REFS,
@@ -214,6 +229,7 @@ enum gotd_imsg_type {
 	/* Sending or receiving a pack file. */
 	GOTD_IMSG_SEND_PACKFILE, /* The server is sending a pack file. */
 	GOTD_IMSG_RECV_PACKFILE, /* The server is receiving a pack file. */
+	GOTD_IMSG_PACKFILE_RECEIVED,
 	GOTD_IMSG_PACKIDX_FILE,  /* Temporary file handle for new pack index. */
 	GOTD_IMSG_PACKFILE_PIPE, /* Pipe to send/receive a pack file stream. */
 	GOTD_IMSG_PACKFILE_PROGRESS, /* Progress reporting. */
@@ -222,6 +238,12 @@ enum gotd_imsg_type {
 	GOTD_IMSG_PACKFILE_INSTALL, /* Received pack file can be installed. */
 	GOTD_IMSG_PACKFILE_DONE, /* Pack file has been sent/received. */
 
+	/* Pack file content verification. */
+	GOTD_IMSG_PACKFILE_GET_CONTENT,
+	GOTD_IMSG_PACKFILE_CONTENT_WRITTEN,
+	GOTD_IMSG_RUN_GOTSYS_CHECK,
+	GOTD_IMSG_PACKFILE_VERIFIED,
+
 	/* Reference updates. */
 	GOTD_IMSG_REF_UPDATES_START, /* Ref updates starting. */
 	GOTD_IMSG_REF_UPDATE_OK, /* Update went OK. */
@@ -229,6 +251,10 @@ enum gotd_imsg_type {
 	GOTD_IMSG_REFS_UPDATED, /* The server processed all ref updates. */
 
 	/* Client connections. */
+	GOTD_IMSG_LISTENER_READY,
+	GOTD_IMSG_LISTEN_SOCKET,
+	GOTD_IMSG_CONNECTION_LIMIT,
+	GOTD_IMSG_REQUEST_TIMEOUT,
 	GOTD_IMSG_DISCONNECT,
 	GOTD_IMSG_CONNECT,
 
@@ -237,11 +263,34 @@ enum gotd_imsg_type {
 	GOTD_IMSG_REPO_CHILD_READY,
 	GOTD_IMSG_CONNECT_REPO_CHILD,
 
+	/* Reloading. */
+	GOTD_IMSG_RELOAD_READY,
+	GOTD_IMSG_RELOAD_SECRETS,
+	GOTD_IMSG_GOTD_CONF,
+	GOTD_IMSG_HALT,
+
 	/* Auth child process. */
+	GOTD_IMSG_AUTH_READY,
+	GOTD_IMSG_AUTH_ACCESS_RULE,
 	GOTD_IMSG_AUTHENTICATE,
 	GOTD_IMSG_ACCESS_GRANTED,
 
+	/* Protected references. */
+	GOTD_IMSG_PROTECTED_TAG_NAMESPACES,
+	GOTD_IMSG_PROTECTED_TAG_NAMESPACES_ELEM,
+	GOTD_IMSG_PROTECTED_BRANCH_NAMESPACES,
+	GOTD_IMSG_PROTECTED_BRANCH_NAMESPACES_ELEM,
+	GOTD_IMSG_PROTECTED_BRANCHES,
+	GOTD_IMSG_PROTECTED_BRANCHES_ELEM,
+
 	/* Notify child process. */
+	GOTD_IMSG_NOTIFIER_READY,
+	GOTD_IMSG_NOTIFICATION_REFS,
+	GOTD_IMSG_NOTIFICATION_REFS_ELEM,
+	GOTD_IMSG_NOTIFICATION_REF_NAMESPACES,
+	GOTD_IMSG_NOTIFICATION_REF_NAMESPACES_ELEM,
+	GOTD_IMSG_NOTIFICATION_TARGET_EMAIL,
+	GOTD_IMSG_NOTIFICATION_TARGET_HTTP,
 	GOTD_IMSG_CONNECT_NOTIFIER,
 	GOTD_IMSG_CONNECT_SESSION,
 	GOTD_IMSG_NOTIFY,
@@ -285,6 +334,7 @@ struct gotd_imsg_info_client {
 	int is_writing;
 	pid_t session_child_pid;
 	pid_t repo_child_pid;
+	time_t time_connected;
 };
 
 /* Structure for GOTD_IMSG_LIST_REFS. */
@@ -376,6 +426,30 @@ struct gotd_imsg_ref_update {
 	/* Followed by name_len data bytes. */
 } __attribute__((__packed__));
 
+/* Structure for GOTD_IMSG_PACKFILE_GET_CONTENT. */
+struct gotd_imsg_packfile_get_content {
+	size_t refname_len;
+	size_t path_len;
+
+	/* Followed by refname_len + path_len data bytes. */
+
+	/* Content file descriptor is passed via imsg. */
+};
+
+/* Structure for GOTD_IMSG_PACKFILE_CONTENT_WRITTEN. */
+struct gotd_imsg_packfile_content_written {
+	/* If zero, the requested reference is not being updated. */
+	int ref_found;
+
+	/*
+	 * If wrote_content is zero, nothing was found at the requested path.
+	 *
+	 * Else, content as it appears in the pack file was written to the
+	 * file descriptor. If this content is empty, the file will be empty.
+	 */
+	int wrote_content;
+};
+
 /* Structure for GOTD_IMSG_REF_UPDATES_START data. */
 struct gotd_imsg_ref_updates_start {
 	int nref_updates;
@@ -419,6 +493,11 @@ struct gotd_imsg_recv_packfile {
 	/* pack destination temp file is sent as a file descriptor */
 };
 
+/* Structure for GOTD_IMSG_PACKFILE_RECEIVED data. */
+struct gotd_imsg_packfile_received {
+	int pack_empty;
+};
+
 /*
  * Structure for GOTD_IMSG_PACKFILE_PROGRESS and
  * GOTD_IMSG_PACKFILE_READY data.
@@ -439,6 +518,13 @@ struct gotd_imsg_packfile_install {
 	uint8_t pack_sha1[SHA1_DIGEST_LENGTH];
 };
 
+/* Structure for GOTD_IMSG_LISTEN_SOCKET data. */
+struct gotd_imsg_listen_socket {
+	size_t nconnection_limits;
+
+	/* listen fd passed via imsg */
+};
+
 /* Structure for GOTD_IMSG_DISCONNECT data. */
 struct gotd_imsg_disconnect {
 	uint32_t client_id;
@@ -456,6 +542,7 @@ struct gotd_imsg_connect {
 
 /* Structure for GOTD_IMSG_CONNECT_REPO_CHILD. */
 struct gotd_imsg_connect_repo_child {
+	char repo_name[NAME_MAX];
 	enum gotd_procid proc_id;
 
 	/* repo child imsg pipe is passed via imsg fd */
@@ -467,6 +554,77 @@ struct gotd_imsg_auth {
 	gid_t egid;
 	int required_auth;
 	uint32_t client_id;
+	char repo_name[NAME_MAX];
+};
+
+/* Structure for GOTD_IMSG_AUTH_ACCESS_RULE. */
+struct gotd_imsg_auth_access_rule {
+	enum gotd_access access;
+	int authorization;
+	size_t identifier_len;
+
+	/* Followed by identifier_len bytes. */
+};
+
+/*
+ * Structure for sending path lists over imsg. Used with:
+ * GOTD_IMSG_PROTECTED_TAG_NAMESPACES
+ * GOTD_IMSG_PROTECTED_BRANCH_NAMESPACES
+ * GOTD_IMSG_PROTECTED_BRANCHES
+ * GOTD_IMSG_NOTIFY_BRANCHES
+ * GOTD_IMSG_NOTIFY_REF_NAMESPACES
+ */
+struct gotd_imsg_pathlist {
+	size_t nelem;
+
+	/* Followed by nelem path list elements. */
+};
+
+/*
+ * Structure for a path list element. Used with:
+ * GOTD_IMSG_PROTECTED_TAG_NAMESPACES_ELEM
+ * GOTD_IMSG_PROTECTED_BRANCH_NAMESPACES_ELEM
+ * GOTD_IMSG_PROTECTED_BRANCHES_ELEM
+ * GOTD_IMSG_NOTIFY_BRANCHES_ELEM
+ * GOTD_IMSG_NOTIFY_REF_NAMESPACES_ELEM
+ */
+struct gotd_imsg_pathlist_elem {
+	size_t path_len;
+	size_t data_len;
+
+	/* Followed by path_len bytes. */
+	/* Followed by data_len bytes. */
+};
+
+/* Structure for GOTD_IMSG_NOTIFICATION_TARGET_EMAIL. */
+struct gotd_imsg_notitfication_target_email {
+	size_t sender_len;
+	size_t recipient_len;
+	size_t responder_len;
+	size_t hostname_len;
+	size_t port_len;
+	size_t repo_name_len;
+
+	/*
+	 * Followed by sender_len + responder_len + responder_len +
+	 * hostname_len + port_len + repo_name_len bytes.
+	 */
+};
+
+/* Structure for GOTD_IMSG_NOTIFICATION_TARGET_HTTP. */
+struct gotd_imsg_notitfication_target_http {
+	int tls;
+	size_t hostname_len;
+	size_t port_len;
+	size_t path_len;
+	size_t auth_len;
+	size_t hmac_len;;
+	size_t repo_name_len;
+
+	/*
+	 * Followed by hostname_len + port_len + path_len + auth_len +
+	 * hmac_len + repo_name_len bytes.
+	 */
 };
 
 /* Structures for GOTD_IMSG_NOTIFY. */
@@ -492,8 +650,8 @@ struct gotd_imsg_notify {
 };
 
 int enter_chroot(const char *);
-int parse_config(const char *, enum gotd_procid, struct gotd_secrets *,
-    struct gotd *);
+int gotd_parse_config(const char *, int, enum gotd_procid,
+    struct gotd_secrets *, struct gotd *);
 struct gotd_repo *gotd_find_repo_by_name(const char *, struct gotd_repolist *);
 struct gotd_repo *gotd_find_repo_by_path(const char *, struct gotd *);
 struct gotd_uid_connection_limit *gotd_find_uid_connection_limit(
@@ -501,10 +659,10 @@ struct gotd_uid_connection_limit *gotd_find_uid_connection_limit(
 int gotd_parseuid(const char *s, uid_t *uid);
 const struct got_error *gotd_parse_url(char **, char **, char **,
     char **, const char *);
+const struct got_error *gotd_conf_new_repo(struct gotd_repo **, const char *);
 
 /* imsg.c */
 const struct got_error *gotd_imsg_flush(struct imsgbuf *);
-const struct got_error *gotd_imsg_recv(struct imsg *, struct imsgbuf *, size_t);
 const struct got_error *gotd_imsg_poll_recv(struct imsg *, struct imsgbuf *,
     size_t);
 const struct got_error *gotd_imsg_recv_error(uint32_t *client_id,
@@ -522,3 +680,11 @@ void gotd_imsg_send_ack(struct got_object_id *, struct imsgbuf *,
     uint32_t, pid_t);
 void gotd_imsg_send_nak(struct got_object_id *, struct imsgbuf *,
     uint32_t, pid_t);
+const struct got_error *gotd_imsg_recv_pathlist(size_t *, struct imsg *);
+const struct got_error *gotd_imsg_recv_pathlist_elem(struct imsg *,
+    struct got_pathlist_head *);
+void gotd_free_notification_target(struct gotd_notification_target *);
+const struct got_error *gotd_imsg_recv_notification_target_email(char **,
+    struct gotd_notification_target **, struct imsg *);
+const struct got_error *gotd_imsg_recv_notification_target_http(char **,
+    struct gotd_notification_target **, struct imsg *);
