@@ -433,6 +433,7 @@ got_get_repo_commits(struct request *c, size_t limit)
 			t->more_id = strdup(repo_commit->commit_id);
 			if (t->more_id == NULL)
 				error = got_error_from_errno("strdup");
+			gotweb_free_repo_commit(repo_commit);
 			goto done;
 		}
 
@@ -599,6 +600,10 @@ got_get_repo_tags(struct request *c, size_t limit)
 
 		if (commit_found == 0 && qs->commit != NULL &&
 		    strncmp(id_str, qs->commit, strlen(id_str)) != 0) {
+			if (commit) {
+				got_object_commit_close(commit);
+				commit = NULL;
+			}
 			TAILQ_REMOVE(&t->repo_tags, new_repo_tag, entry);
 			gotweb_free_repo_tag(new_repo_tag);
 			continue;
@@ -786,6 +791,7 @@ got_open_blob_for_output(struct got_blob_object **blob, int *fd,
 	struct got_repository *repo = c->t->repo;
 	struct got_commit_object *commit = NULL;
 	struct got_object_id *commit_id = NULL;
+	struct got_object_id *blob_id = NULL;
 	struct got_reflist_head refs;
 	char *path = NULL, *in_repo_path = NULL;
 	int obj_type;
@@ -823,16 +829,16 @@ got_open_blob_for_output(struct got_blob_object **blob, int *fd,
 	if (error)
 		goto done;
 
-	error = got_object_id_by_path(&commit_id, repo, commit, in_repo_path);
+	error = got_object_id_by_path(&blob_id, repo, commit, in_repo_path);
 	if (error)
 		goto done;
 
-	if (commit_id == NULL) {
+	if (blob_id == NULL) {
 		error = got_error(GOT_ERR_NO_OBJ);
 		goto done;
 	}
 
-	error = got_object_get_type(&obj_type, repo, commit_id);
+	error = got_object_get_type(&obj_type, repo, blob_id);
 	if (error)
 		goto done;
 
@@ -845,7 +851,7 @@ got_open_blob_for_output(struct got_blob_object **blob, int *fd,
 	if (error)
 		goto done;
 
-	error = got_object_open_as_blob(blob, repo, commit_id, BUF, *fd);
+	error = got_object_open_as_blob(blob, repo, blob_id, BUF, *fd);
 	if (error)
 		goto done;
 
@@ -869,6 +875,7 @@ got_open_blob_for_output(struct got_blob_object **blob, int *fd,
 	got_ref_list_free(&refs);
 	free(in_repo_path);
 	free(commit_id);
+	free(blob_id);
 	free(path);
 	return error;
 }
@@ -1100,7 +1107,7 @@ got_output_file_blame(struct request *c, got_render_blame_line_cb cb)
 		goto done;
 
 	error = got_blame(in_repo_path, commit_id, repo,
-	    GOT_DIFF_ALGORITHM_MYERS, got_gotweb_blame_cb, &bca, NULL, NULL,
+	    GOT_DIFF_ALGORITHM_PATIENCE, got_gotweb_blame_cb, &bca, NULL, NULL,
 	    fd1, fd2, f1, f2);
 
  done:
@@ -1205,18 +1212,18 @@ got_open_diff_for_output(FILE **fp, struct request *c)
 	switch (obj_type) {
 	case GOT_OBJ_TYPE_BLOB:
 		error = got_diff_objects_as_blobs(NULL, NULL, f1, f2, fd1, fd2,
-		     id1, id2, NULL, NULL, GOT_DIFF_ALGORITHM_MYERS, 3, 0, 0,
-		     NULL, repo, f3);
+		     id1, id2, NULL, NULL, GOT_DIFF_ALGORITHM_PATIENCE,
+		     3, 0, 0, NULL, repo, f3);
 		break;
 	case GOT_OBJ_TYPE_TREE:
 		error = got_diff_objects_as_trees(NULL, NULL, f1, f2, fd1, fd2,
-		    id1, id2, NULL, "", "",  GOT_DIFF_ALGORITHM_MYERS, 3, 0, 0,
-		    NULL, repo, f3);
+		    id1, id2, NULL, "", "", GOT_DIFF_ALGORITHM_PATIENCE,
+		    3, 0, 0, NULL, repo, f3);
 		break;
 	case GOT_OBJ_TYPE_COMMIT:
 		error = got_diff_objects_as_commits(NULL, NULL, f1, f2, fd1,
-		    fd2, id1, id2, NULL,  GOT_DIFF_ALGORITHM_MYERS, 3, 0, 0,
-		    NULL, repo, f3);
+		    fd2, id1, id2, NULL,  GOT_DIFF_ALGORITHM_PATIENCE,
+		    3, 0, 0, NULL, repo, f3);
 		break;
 	default:
 		error = got_error(GOT_ERR_OBJ_TYPE);
