@@ -277,6 +277,14 @@ got_pack_paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 				err = got_pack_paint_commit(qid, COLOR_SKIP);
 				if (err)
 					goto done;
+				err = got_object_idset_add(skip, &qid->id,
+				    NULL);
+				if (err)
+					goto done;
+				err = got_pack_repaint_parent_commits(&qid->id,
+				    COLOR_SKIP, skip, skip, repo);
+				if (err)
+					goto done;
 			} else
 				(*ncolored)++;
 			err = got_object_idset_add(keep, &qid->id, NULL);
@@ -288,6 +296,14 @@ got_pack_paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 				err = got_pack_paint_commit(qid, COLOR_SKIP);
 				if (err)
 					goto done;
+				err = got_object_idset_add(skip, &qid->id,
+				    NULL);
+				if (err)
+					goto done;
+				err = got_pack_repaint_parent_commits(&qid->id,
+				    COLOR_SKIP, skip, skip, repo);
+				if (err)
+					goto done;
 			} else
 				(*ncolored)++;
 			err = got_object_idset_add(drop, &qid->id, NULL);
@@ -327,7 +343,7 @@ got_pack_paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 				err = got_pack_queue_commit_id(ids, &pid->id,
 				    color, repo);
 				if (err)
-					break;
+					goto done;
 				nqueued++;
 				if (color == COLOR_SKIP)
 					nskip++;
@@ -342,7 +358,7 @@ got_pack_paint_commits(int *ncolored, struct got_object_id_queue *ids, int nids,
 			 */
 			if (packidx == NULL) {
 				err = got_pack_find_pack_for_commit_painting(
-				    &packidx, ids, nqueued, repo);
+				    &packidx, ids, repo);
 				if (err)
 					goto done;
 			}
