@@ -81,7 +81,7 @@ match_remote_ref(struct got_pathlist_head *have_refs,
 	 * we should use a flag instead */
 	memset(my_id, 0, sizeof(*my_id));
 
-	TAILQ_FOREACH(pe, have_refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, have_refs) {
 		struct got_object_id *id = pe->data;
 		if (strcmp(pe->path, refname) == 0) {
 			memcpy(my_id, id, sizeof(*my_id));
@@ -191,13 +191,16 @@ fetch_progress(struct imsgbuf *ibuf, const char *buf, size_t len)
 	if (len > MAX_IMSGSIZE - IMSG_HEADER_SIZE)
 		len = MAX_IMSGSIZE - IMSG_HEADER_SIZE;
 
-	/* Only allow printable ASCII. */
+	/*
+	 * We only allow printable ASCII for safety reasons and silently
+	 * ignore non-ASCII messages, which may be somewhat legitimately
+	 * caused by Git servers running in a non-English language locale.
+	 */
 	for (i = 0; i < len; i++) {
 		if (isprint((unsigned char)buf[i]) ||
 		    isspace((unsigned char)buf[i]))
 			continue;
-		return got_error_msg(GOT_ERR_BAD_PACKET,
-		    "non-printable progress message received from server");
+		return NULL;
 	}
 
 	return send_fetch_server_progress(ibuf, buf, len);
@@ -227,7 +230,7 @@ send_fetch_symrefs(struct imsgbuf *ibuf, struct got_pathlist_head *symrefs)
 	struct got_pathlist_entry *pe;
 
 	len = sizeof(struct got_imsg_fetch_symrefs);
-	TAILQ_FOREACH(pe, symrefs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, symrefs) {
 		const char *target = pe->data;
 		len += sizeof(struct got_imsg_fetch_symref) +
 		    pe->path_len + strlen(target);
@@ -245,7 +248,7 @@ send_fetch_symrefs(struct imsgbuf *ibuf, struct got_pathlist_head *symrefs)
 	if (imsg_add(wbuf, &nsymrefs, sizeof(nsymrefs)) == -1)
 		return got_error_from_errno("imsg_add FETCH_SYMREFS");
 
-	TAILQ_FOREACH(pe, symrefs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, symrefs) {
 		const char *name = pe->path;
 		size_t name_len = pe->path_len;
 		const char *target = pe->data;
@@ -356,7 +359,7 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 	ssize_t w;
 	struct got_ratelimit rl;
 
-	TAILQ_INIT(&symrefs);
+	RB_INIT(&symrefs);
 	got_hash_init(&ctx, GOT_HASH_SHA1);
 	got_ratelimit_init(&rl, 0, 500);
 
@@ -422,7 +425,7 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 				goto done;
 			is_firstpkt = 0;
 			if (!fetch_all_branches) {
-				TAILQ_FOREACH(pe, &symrefs, entry) {
+				RB_FOREACH(pe, got_pathlist_head, &symrefs) {
 					const char *name = pe->path;
 					const char *symref_target = pe->data;
 					if (strcmp(name, GOT_REF_HEAD) != 0)
@@ -466,7 +469,7 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 				found_branch = 1;
 				continue;
 			}
-			TAILQ_FOREACH(pe, wanted_branches, entry) {
+			RB_FOREACH(pe, got_pathlist_head, wanted_branches) {
 				if (match_branch(refname, pe->path))
 					break;
 			}
@@ -483,7 +486,7 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 				    getprogname(), refname);
 			}
 		} else {
-			TAILQ_FOREACH(pe, wanted_refs, entry) {
+			RB_FOREACH(pe, got_pathlist_head, wanted_refs) {
 				if (match_wanted_ref(refname, pe->path))
 					break;
 			}
@@ -531,7 +534,7 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 		struct got_pathlist_entry *pe;
 		static char msg[PATH_MAX + 33];
 
-		pe = TAILQ_FIRST(wanted_branches);
+		pe = RB_MIN(got_pathlist_head, wanted_branches);
 		if (pe) {
 			snprintf(msg, sizeof(msg),
 			    "branch \"%s\" not found on server", pe->path);
@@ -539,7 +542,7 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 			goto done;
 		}
 
-		pe = TAILQ_FIRST(wanted_refs);
+		pe = RB_MIN(got_pathlist_head, wanted_refs);
 		if (pe) {
 			snprintf(msg, sizeof(msg),
 			    "reference \"%s\" not found on server", pe->path);
@@ -554,7 +557,7 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 	for (i = 0; i < nref; i++) {
 		if (got_object_id_cmp(&have[i], &want[i]) == 0)
 			continue;
-		got_sha1_digest_to_str(want[i].sha1, hashstr, sizeof(hashstr));
+		got_object_id_hex(&want[i], hashstr, sizeof(hashstr));
 		n = snprintf(buf, sizeof(buf), "want %s%s\n", hashstr,
 		    sent_my_capabilites || my_capabilities == NULL ?
 		    "" : my_capabilities);
@@ -575,9 +578,9 @@ fetch_pack(int fd, int packfd, uint8_t *pack_sha1,
 	if (nwant == 0)
 		goto done;
 
-	TAILQ_FOREACH(pe, have_refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, have_refs) {
 		struct got_object_id *id = pe->data;
-		got_sha1_digest_to_str(id->sha1, hashstr, sizeof(hashstr));
+		got_object_id_hex(id, hashstr, sizeof(hashstr));
 		n = snprintf(buf, sizeof(buf), "have %s\n", hashstr);
 		if (n < 0 || (size_t)n >= sizeof(buf)) {
 			err = got_error(GOT_ERR_NO_SPACE);
@@ -864,6 +867,7 @@ main(int argc, char **argv)
 	struct got_imsg_fetch_wanted_branch wbranch;
 	struct got_imsg_fetch_wanted_ref wref;
 	size_t datalen, i;
+	struct got_pathlist_entry *new;
 	char *remote_head = NULL, *worktree_branch = NULL;
 #if 0
 	static int attached;
@@ -871,16 +875,22 @@ main(int argc, char **argv)
 		sleep (1);
 #endif
 
-	TAILQ_INIT(&have_refs);
-	TAILQ_INIT(&wanted_branches);
-	TAILQ_INIT(&wanted_refs);
+	RB_INIT(&have_refs);
+	RB_INIT(&wanted_branches);
+	RB_INIT(&wanted_refs);
+
+	if (imsgbuf_init(&ibuf, GOT_IMSG_FD_CHILD) == -1) {
+		warn("imsgbuf_init");
+		return 1;
+	}
+	imsgbuf_allow_fdpass(&ibuf);
 
-	imsg_init(&ibuf, GOT_IMSG_FD_CHILD);
 #ifndef PROFILE
 	/* revoke access to most system calls */
 	if (pledge("stdio recvfd", NULL) == -1) {
 		err = got_error_from_errno("pledge");
 		got_privsep_send_error(&ibuf, err);
+		imsgbuf_clear(&ibuf);
 		return 1;
 	}
 
@@ -984,11 +994,12 @@ main(int argc, char **argv)
 			goto done;
 		}
 		memcpy(id, &href.id, sizeof(*id));
-		err = got_pathlist_append(&have_refs, refname, id);
-		if (err) {
+		err = got_pathlist_insert(&new, &have_refs, refname, id);
+		if (err || new == NULL) {
 			free(refname);
 			free(id);
-			goto done;
+			if (err)
+				goto done;
 		}
 
 		imsg_free(&imsg);
@@ -1026,10 +1037,11 @@ main(int argc, char **argv)
 			goto done;
 		}
 
-		err = got_pathlist_append(&wanted_branches, refname, NULL);
-		if (err) {
+		err = got_pathlist_insert(&new, &wanted_branches, refname, NULL);
+		if (err || new == NULL) {
 			free(refname);
-			goto done;
+			if (err)
+				goto done;
 		}
 
 		imsg_free(&imsg);
@@ -1066,7 +1078,7 @@ main(int argc, char **argv)
 			goto done;
 		}
 
-		err = got_pathlist_append(&wanted_refs, refname, NULL);
+		err = got_pathlist_insert(NULL, &wanted_refs, refname, NULL);
 		if (err) {
 			free(refname);
 			goto done;
@@ -1113,7 +1125,10 @@ done:
 	if (err != NULL) {
 		fprintf(stderr, "%s: %s\n", getprogname(), err->msg);
 		got_privsep_send_error(&ibuf, err);
+		imsgbuf_clear(&ibuf);
+		exit(1);
 	}
 
+	imsgbuf_clear(&ibuf);
 	exit(0);
 }
